# -*- coding: utf-8 -*-
# @Author  : BoPo
# @Time    : 2022/2/17 14:49
# @Function:
from datetime import datetime
from pathlib import Path

from pytdx.reader import CustomerBlockReader

from .. import utils
from ..consts import TYPE_GROUP, TYPE_FLATS
from ..logger import logger
from ..utils import get_stock_market


class Customize:
    items: dict = {}

    def __init__(self, tdxdir=None):
        self.vipdoc = Path(tdxdir, 'T0002', 'blocknew')
        self.tdxdir = str(tdxdir)

    def create(self, name: str = None, symbol: list = None, **kwargs):
        return utils.block_new(self.tdxdir, name=name, symbol=symbol, **kwargs)

    def remove(self, name: str):
        # 板块数据
        block_data = self.search()
        block_file = Path(self.vipdoc) / 'blocknew.cfg'
        block_temp = block_data[block_data.blockname == name]
        block_type = ''

        if block_temp.block_type.to_list():
            block_type = list(set(block_temp.block_type.to_list()))[0]
            [Path(self.vipdoc, f"{x}.blk").unlink() for x in block_temp.block_type.to_list() if Path(self.vipdoc, f"{x}.blk").is_file()]

        block_data = Path(block_file).read_bytes().decode(encoding='gb2312')
        data = name + ((50 - len(name.encode('gbk', 'ignore'))) * '\x00')
        data += block_type + ((70 - len(block_type.encode('gbk', 'ignore'))) * '\x00')
        data = block_data.replace(data, '')
        data = bytes(data.encode('gbk', 'ignore'))

        return Path(block_file).write_bytes(data)

    def search(self, name: str = None, group=None):
        types_ = TYPE_GROUP if group else TYPE_FLATS

        if name:
            result = CustomerBlockReader().get_df(str(self.vipdoc), TYPE_GROUP)
            result = result.loc[result.blockname == name]

            if result.empty:
                return None

            result = result.code_list.values
            result = list(set(result[0].split(',')))
        else:
            result = CustomerBlockReader().get_df(str(self.vipdoc), types_)

        return result

    def update(self, name, symbol=None, overflow=False):
        """
        修改自定义板块内容

        :param name: 板块名称
        :param symbol: 股票代码, ['600036','600016']
        :param overflow:
        """

        # 板块路径
        block_path = self.vipdoc
        block_code = list(symbol)

        # 板块数据
        block_data = self.search()
        block_temp = block_data[block_data.blockname == name]

        # 对于名称空的情况, 直接创建写入
        if block_temp.empty:
            logger.debug('block_temp is empty {}', block_temp.empty)
            return utils.block_new(self.tdxdir, name=name, symbol=list(set(symbol)))

        # 覆盖情况
        if not overflow:
            block_code += block_temp.code.to_list()

        # 取 blk 文件名, block_type 不为空
        if block_temp.block_type.to_list():
            block_type = list(set(block_temp.block_type.to_list()))[0]
            logger.debug('发现板块文件: {}', block_type)
        else:
            # block_type 为空的话
            block_type = datetime.now().strftime('%Y%m%d%H%M%S')
            logger.debug('板块文件找不到: {}', block_type)

        # 去重股票代码
        block_code = list(set(block_code))
        logger.debug('证券代码: {}', block_code)

        # 股票代码逗号隔开拼字符串
        block_code = '\n'.join([f"{get_stock_market(s)}{s}" for s in block_code])

        # 写入 blk 文件
        logger.debug("写入文件 : {}", Path(block_path, f"{block_type}.blk"))
        return Path(block_path, f"{block_type}.blk").write_text(block_code, encoding='gb2312')



import sys

from loguru import logger

logger.remove()


def getLogger(quiet=None, verbose=None) -> logger:  # noqa
    level = ('INFO', 'DEBUG')[bool(verbose)]
    logger.remove()

    quiet or logger.add(sys.stderr, level=level)

    return logger


def reset(verbose: int = 0, **kwargs) -> logger:  # noqa
    """
    重置 logger 等级函数

    :param verbose: 等级级别 0 - 3
    :param kwargs:
    :return:
    """

    levels = ['WARNING', 'INFO', 'DEBUG', 'TRACE']
    level0 = levels[-1] if verbose > len(levels) else levels[verbose]

    logger.remove()
    logger.add(sys.stdout, level=level0)

    return logger


def setup(verbose: int = 0, **kwargs) -> logger:  # noqa
    return reset(verbose=verbose, **kwargs)

def config(verbose: int = 0, **kwargs) -> logger:  # noqa
    return reset(verbose=verbose, **kwargs)



import time
from datetime import datetime
from pathlib import Path

from loguru import logger
from pytdx.reader import CustomerBlockReader, BlockReader

from mootdx.consts import TYPE_GROUP, TYPE_FLATS
from mootdx.utils import get_stock_market


def _blocknew(tdxdir: str = None, name: str = None, symbol: list = None, blk_file: str = None, **kwargs):
    """
    自定义模块写入函数

    :param tdxdir: tdx 路径
    :param name: 自定义板块名称
    :param symbol: 自定义板块股票代码集合
    :param blk_file: 自定义板块股票代码集合文件
    :return: bool
    """

    if not tdxdir:
        logger.error(f'通达信路径不存在或者空: {tdxdir}')
        return False

    # 自定义板块名称未传入则自动按时间生成名称
    if not name:
        name = datetime.now().strftime('%Y%m%d%H%M%S')

    # 按时间生成 blk 文件名
    blk_file = blk_file if blk_file else str(time.time_ns())

    vipdoc = Path(tdxdir, 'T0002', 'blocknew')
    symbol = list(set(symbol))

    # 判断目录是否存在
    if not Path(vipdoc).is_dir():
        logger.error(f'自定义板块目录错误: {vipdoc}')
        return False

    block_file = Path(vipdoc) / 'blocknew.cfg'

    # 文件不存在就创建
    if not block_file.exists():
        block_file.write_text('')

    # 判断名字是否重名
    with open(block_file, 'rb') as fp:
        names = fp.read().decode('gbk', 'ignore')
        names = names.split('\x00')
        names = [x for x in names if x != '']
        names = [v for i, v in enumerate(names) if i % 2 == 0]

        if name in names:
            # todo symbol 不空则合并, 空则删除
            logger.error('自定义板块名称重复.')
            raise Exception('自定义板块名称重复.')

    # 写 blk 文件
    with open(f'{vipdoc}/{blk_file}.blk', 'w') as fp:
        fp.write('\n'.join([f"{get_stock_market(s)}{s}" for s in symbol]))

    # 写 blocknew.cfg 文件
    with open(block_file, 'ab') as fp:
        data = name + ((50 - len(name.encode('gbk', 'ignore'))) * '\x00')
        data += blk_file + ((70 - len(blk_file.encode('gbk', 'ignore'))) * '\x00')
        data = bytes(data.encode('gbk', 'ignore'))
        fp.write(data)

    return True


def blocknew(tdxdir: str = None, name: str = None, symbol: list = None, group=False, **kwargs):
    """
    自定义板块数据操作

    提示: name 和 symbol 全为空则为读取，否则写入操作
    参考: http://blog.sina.com.cn/s/blog_623d2d280102vt8y.html

    :param name: 自定义板块名称
    :param symbol: 自定义板块股票列表, 类型 list
    :param group:
    :return: pd.dataFrame or Bool
    """

    if name and symbol:
        return _blocknew(tdxdir, name=name, symbol=symbol, **kwargs)

    vipdoc = Path(tdxdir, 'T0002', 'blocknew')
    types_ = TYPE_GROUP if group else TYPE_FLATS

    if not vipdoc.is_dir():
        return None

    if name:
        result = CustomerBlockReader().get_df(str(vipdoc), TYPE_GROUP)
        result = result.loc[result.blockname == name].code_list.values
        result = list(set(result[0].split(',')))
    else:
        result = CustomerBlockReader().get_df(str(vipdoc), types_)

    return result


def block(tdxdir, symbol='', group=False, **kwargs):
    """
    获取板块数据

    参考: http://blog.sina.com.cn/s/blog_623d2d280102vt8y.html

    :param symbol:  板块文件
    :param group:   分组解析
    :return: pd.dataFrame or None
    """

    suffix = Path(symbol).suffix
    suffix = suffix if suffix else 'dat'

    symbol = symbol.replace(suffix, '')
    suffix = suffix.strip('.')

    if 'incon' in symbol:
        vipdoc = Path(tdxdir) / f'{symbol}.{suffix}'
    else:
        vipdoc = Path(tdxdir) / 'T0002' / 'hq_cache' / f'{symbol}.{suffix}'

    vipdoc.exists() or logger.debug(f'文件不存在: {vipdoc}')

    types_ = TYPE_GROUP if group else TYPE_FLATS

    if kwargs.get('debug'):
        return str(vipdoc)

    return BlockReader().get_df(str(vipdoc), types_) if vipdoc.exists() else None



import os
import random
import shutil
import tempfile
from pathlib import Path
from struct import calcsize
from struct import unpack

import pandas as pd
from pytdx.hq import TdxHq_API

from .base import BaseFinancial
from ..logger import logger
from ..logger import logger


class FinancialReader(object):
    @staticmethod
    def to_data(filename):
        """
        读取历史财务数据文件，并返回pandas结果 ， 类似 `gpcw20171231.zip` 格式，具体字段含义参考

        https://github.com/rainx/pytdx/issues/133

        :param filename: 数据文件地址， 数据文件类型可以为 .zip 文件，也可以为解压后的 .dat, 可以不写扩展名. 程序自动识别
        :return: pandas DataFrame 格式的历史财务数据
        """

        crawler = Financial()

        with open(filename, 'rb') as fp:
            data = crawler.parse(download_file=fp)

        return crawler.to_df(data)


class FinancialList(BaseFinancial):

    def content(self, report_hook=None, downdir=None, proxies=None, chunk_size=1024 * 50, *args, **kwargs):
        """
        解析财务文件

        :param report_hook: 钩子回调函数
        :param downdir: 要解析的文件夹
        :param proxies:
        :param chunk_size:
        :param args:
        :param kwargs:
        :return:
        """

        tmp = tempfile.NamedTemporaryFile(delete=True)

        api = TdxHq_API(**kwargs)
        api.need_setup = False

        with api.connect(*self.bestip):
            content = api.get_report_file_by_size('tdxfin/gpcw.txt')
            download_file = (open(downdir, 'wb') if downdir else tmp)
            download_file.write(content)
            download_file.seek(0)

            return download_file

    def parse(self, download_file, *args, **kwargs):
        """
        解析财务文件

        :param download_file:
        :param args:
        :param kwargs:
        :return:
        """

        with download_file:
            content = download_file.read()
            content = content.decode('utf-8')

        def l2d(i):
            return {'filename': i[0], 'hash': i[1], 'filesize': int(i[2])}

        if content:
            content = content.strip().split('\n')
            return [l2d(i) for i in [line.strip().split(',') for line in content]]

        return None


class Financial(BaseFinancial):

    def content(self, report_hook=None, downdir=None, proxies=None, chunk_size=51200, *args, **kwargs):
        """
        解析财务文件

        :param report_hook: 钩子回调函数
        :param downdir: 要解析的文件夹
        :param proxies: 代理配置
        :param chunk_size:
        :param args:
        :param kwargs:
        :return:
        """

        filename = kwargs.get('filename')
        downfile = str(Path(downdir) / filename)
        filesize = kwargs.get('filesize') if kwargs.get('filesize') else 0

        logger.info('{}: start download...', filename)

        if not filename:
            raise Exception('Param filename is not set')

        api = TdxHq_API()
        api.need_setup = False

        with api.connect(*self.bestip):
            content = api.get_report_file_by_size(f'tdxfin/{filename}', filesize=filesize, reporthook=report_hook)
            download_file = (open(downfile, 'wb') if downfile else tempfile.NamedTemporaryFile(delete=True))
            download_file.write(content)
            download_file.seek(0)

            del content

            logger.info('{}: done', filename)

            return download_file

    def parse(self, download_file, *args, **kwargs):
        """
        解析财务文件

        :param download_file: 要解析的文件
        :param args:
        :param kwargs:
        :return:
        """

        header_pack_format = '<1hI1H3L'
        tmpdir = tempfile.gettempdir()

        if download_file.name.endswith('.zip'):
            tmpdir_root = tempfile.gettempdir()
            random_sums = str(random.randint(0, 1000000))
            subdir_name = f'mootdx_{random_sums}'

            tmpdir = Path(tmpdir_root, subdir_name)
            shutil.rmtree(tmpdir, ignore_errors=True)
            Path(tmpdir).mkdir(parents=True)

            shutil.unpack_archive(download_file.name, extract_dir=tmpdir)

            # only one file endswith .dat should be in zip archives
            datfile = None

            for _file in os.listdir(tmpdir):
                if _file.endswith('.dat'):
                    datfile = open(Path(tmpdir, _file), 'rb')

            if datfile is None:
                raise Exception('no dat file found in zip archive')

        elif download_file.name.endswith('.dat'):
            datfile = download_file
        else:
            return None

        header_size = calcsize(header_pack_format)
        stock_item_size = calcsize('<6s1c1L')
        data_header = datfile.read(header_size)
        stock_header = unpack(header_pack_format, data_header)

        max_count = stock_header[2]

        report_date = stock_header[1]
        report_size = stock_header[4]

        report_fields_count = int(report_size / 4)
        report_pack_format = '<{}f'.format(report_fields_count)

        results = []

        for stock_idx in range(0, max_count):
            datfile.seek(header_size + stock_idx * calcsize('<6s1c1L'))
            si = datfile.read(stock_item_size)
            stock_item = unpack('<6s1c1L', si)
            code = stock_item[0].decode('utf-8')
            datfile.seek(stock_item[2])

            info_data = datfile.read(calcsize(report_pack_format))
            cw_info = unpack(report_pack_format, info_data)
            one_record = (code, report_date) + cw_info
            results.append(one_record)

        if download_file.name.endswith('.zip'):
            datfile.close()
            shutil.rmtree(tmpdir, ignore_errors=True)

        download_file.close()

        return results

    @staticmethod
    def to_df(data):
        """
        转换数据为 pandas DataFrame 格式
        :param data: 要转换的数据
        :return: DataFrame
        """

        if len(data) == 0 or len(data[0]) == 0:
            return None

        length = len(data[0]) - 1
        column = ['code', 'report_date']

        for i in range(1, length):
            column.append('col' + str(i))

        df = pd.DataFrame(data=data, columns=column)
        df.set_index('code', inplace=True)

        logger.debug(df)

        return df



import hashlib
from pathlib import Path
from struct import calcsize
from struct import unpack

import pandas as pd
from pandas import DataFrame
from tqdm import tqdm

from mootdx.consts import MARKET_SH, MARKET_BJ
from mootdx.consts import MARKET_SZ
from mootdx.logger import logger
from mootdx.utils.adjust import to_adjust


def get_stock_markets(symbols=None):
    results = []

    assert isinstance(symbols, list), 'stock code need list type'

    if isinstance(symbols, list):
        for symbol in symbols:
            results.append([get_stock_market(symbol, string=False), symbol.strip('sh').strip('sz')])

    return results


def get_stock_market(symbol='', string=False):
    """ 判断股票ID对应的证券市场匹配规则

    ['50', '51', '60', '90', '110'] 为 sh
    ['00', '12'，'13', '18', '15', '16', '18', '20', '30', '39', '115'] 为 sz
    ['5', '6', '9'] 开头的为 sh， 其余为 sz

    :param string: False 返回市场ID，否则市场缩写名称
    :param symbol: 股票ID, 若以 'sz', 'sh' 开头直接返回对应类型，否则使用内置规则判断
    :return 'sh' or 'sz'
    """

    assert isinstance(symbol, str), 'stock code need str type'

    market = 'sh'

    if symbol.startswith(('sh', 'sz', 'SH', 'SZ')):
        market = symbol[:2].lower()

    elif symbol.startswith(('50', '51', '60', '68', '90', '110', '113', '132', '204')):
        market = 'sh'

    elif symbol.startswith(('00', '12', '13', '18', '15', '16', '18', '20', '30', '39', '115', '1318')):
        market = 'sz'

    elif symbol.startswith(('5', '6', '9', '7')):
        market = 'sh'

    elif symbol.startswith(('4', '8')):
        market = 'bj'

    if string is False:

        if market == 'sh':
            market = MARKET_SH

        if market == 'sz':
            market = MARKET_SZ

        if market == 'bj':
            market = MARKET_BJ

    logger.debug(f'market=>{market}')
    return market


def gpcw(filepath):
    cw_file = open(filepath, 'rb')

    header_size = calcsize('<3h1H3L')
    stock_item_size = calcsize('<6s1c1L')

    data_header = cw_file.read(header_size)
    stock_header = unpack('<3h1H3L', data_header)

    max_count = stock_header[3]

    for idx in range(0, max_count):
        cw_file.seek(header_size + idx * calcsize('<6s1c1L'))
        si = cw_file.read(stock_item_size)
        stock_item = unpack('<6s1c1L', si)
        code = stock_item[0].decode()
        foa = stock_item[2]
        cw_file.seek(foa)

        info_data = cw_file.read(calcsize('<264f'))
        cw_info = unpack('<264f', info_data)

        logger.debug(f'{code}, {cw_info}')
        return code, cw_info


def md5sum(downfile):
    """
    文件的 md5 哈希值

    :param downfile: 文件路径
    :return: mixed
    """

    try:
        md5_l = hashlib.md5()
        md5_l.update(Path(downfile).read_bytes())
        return md5_l.hexdigest()
    except (IOError, FileNotFoundError) as e:
        logger.error(f'无法读取文件: {downfile}')
        logger.debug(e)
        return None


def to_data(v, **kwargs):
    """
    数值转换为 pd.DataFrame

    :param v: mixed
    :return: pd.DataFrame
    """

    symbol = kwargs.get('symbol')
    adjust = kwargs.get('adjust', None)

    if adjust in ['01', 'qfq', 'before']:
        adjust = 'qfq'
    elif adjust in ['02', 'hfq', 'after']:
        adjust = 'hfq'
    else:
        adjust = None

    # 空值
    if not v:
        return pd.DataFrame(data=[])

    # DataFrame
    if isinstance(v, DataFrame):
        result = v

    # 列表
    elif isinstance(v, list):
        result = pd.DataFrame(data=v) if len(v) else None

    # 字典
    elif isinstance(v, dict):
        result = pd.DataFrame(data=[v])

    # 空值
    else:
        result = pd.DataFrame(data=[])

    if adjust and adjust in ['qfq', 'hfq'] and symbol:
        from mootdx.utils.adjust import fq_factor
        result = to_adjust(result, symbol=symbol, adjust=adjust)

    if "datetime" in result.columns:
        result.index = pd.to_datetime(result.datetime)
    elif "date" in result.columns:
        result.index = pd.to_datetime(result.date)

    if "vol" in result.columns:
        result['volume'] = result.vol

    return result


def to_file(df, filename=None):
    """
    根据扩展名输出文件

    :param df: pd.DataFrame
    :param filename: 要输出的文件，支持 csv, xlsx, xls, json, h5
    :return: bool
    """
    if filename is None or df is None:
        return None

    path_name = Path(filename).parent
    extension = Path(filename).suffix

    # 目录不存在创建目录
    Path(path_name).is_dir() or Path(path_name).mkdir(parents=True)

    # methods = {'to_json': ['.json']}
    # method = [k for k, v in methods if extension in v][0]
    # getattr(pd, method)(filename)

    if extension == '.csv':
        return df.to_csv(filename, encoding='utf-8', index=False)

    if extension == '.xlsx' or extension == '.xls':
        # openpyxl, xlwt
        return df.to_excel(filename, index=False)

    if extension == '.h5':
        # tables
        return df.to_hdf(filename, 'df', index=False)

    if extension == '.json':
        return df.to_json(filename, orient='records')

    return None


class TqdmUpTo(tqdm):
    """
    Provides `update_to(n)` which uses `tqdm.update(delta_n)`.

    """

    total: object = 0

    def update_to(self, downloaded=0, total_size=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if total_size is not None:
            self.total = total_size

        self.update(downloaded - self.n)  # will also set self.n = b * bsize


def get_config_path(config='config.json'):
    """
    获取配置文件路径

    :param config: 配置文件名称
    :return: filename
    """
    filename = Path.home() / '.mootdx' / config
    pathname = Path(filename).parent

    Path(pathname).exists() or Path(pathname).mkdir(parents=True)
    # Path(filename).exists() or Path(filename).write_text('None')

    return str(filename)


def get_frequency(frequency) -> int:
    FREQUENCY = ['5m', '15m', '30m', '1h', 'day', 'week', 'mon', '1m', '1m', 'day', '3mon', 'year']

    try:
        if isinstance(frequency, str):
            frequency = FREQUENCY.index(frequency)
    except ValueError:
        frequency = 0

    return frequency



from abc import ABC
from pathlib import Path

from pytdx.reader import TdxExHqDailyBarReader
from pytdx.reader import TdxLCMinBarReader
from pytdx.reader import TdxMinBarReader

from mootdx import block
from mootdx.contrib.compat import MooTdxDailyBarReader
from mootdx.utils import get_stock_market


# 股票市场
class Reader(object):
    @staticmethod
    def factory(market='std', **kwargs):
        """
        Reader 工厂方法

        :param market: std 标准市场, ext 扩展市场
        :param kwargs: 可变参数
        :return:
        """

        if market == 'ext':
            return ExtReader(**kwargs)

        return StdReader(**kwargs)


class ReaderBase(ABC):
    """股票市场"""

    # 默认通达信安装目录
    tdxdir = 'C:/new_tdx'

    def __init__(self, tdxdir=None):
        """
        构造函数

        :param tdxdir: 通达信安装目录
        """

        if not Path(tdxdir).is_dir():
            raise Exception('tdxdir 目录不存在')

        self.tdxdir = tdxdir

    def find_path(self, symbol=None, subdir='lday', suffix=None, **kwargs):
        """
        自动匹配文件路径，辅助函数

        :param symbol:
        :param subdir:
        :param suffix:
        :return: pd.dataFrame or None
        """

        # 判断市场, 带#扩展市场
        if '#' in symbol:
            market = 'ds'
        else:
            # 判断是sh还是sz
            market = get_stock_market(symbol, True)

        # 判断前缀(市场是sh和sz重置前缀)
        if market.lower() in ['sh', 'sz']:
            symbol = market + symbol.lower().replace(market, '')

        # 判断后缀
        suffix = suffix if isinstance(suffix, list) else [suffix]

        # 调试使用
        if kwargs.get('debug'):
            return market, symbol, suffix

        # 遍历扩展名
        for ex_ in suffix:
            ex_ = ex_.strip('.')
            vipdoc = Path(self.tdxdir) / 'vipdoc' / market / subdir / f'{symbol}.{ex_}'

            if Path(vipdoc).exists():
                return vipdoc

        return None


class StdReader(ReaderBase):
    """股票市场"""

    def daily(self, symbol=None):
        """
        获取日线数据

        :param symbol: 证券代码
        :return: pd.dataFrame or None
        """
        reader = MooTdxDailyBarReader()
        vipdoc = self.find_path(symbol=symbol, subdir='lday', suffix='day')

        return reader.get_df(str(vipdoc)) if vipdoc else None

    def minute(self, symbol=None, suffix=1):
        """
        获取1, 5分钟线

        :param suffix: 文件前缀
        :param symbol: 证券代码
        :return: pd.dataFrame or None
        """
        subdir = 'fzline' if str(suffix) == '5' else 'minline'
        suffix = ['lc5', '5'] if str(suffix) == '5' else ['lc1', '1']
        symbol = self.find_path(symbol, subdir=subdir, suffix=suffix)

        if symbol is not None:
            reader = (TdxMinBarReader() if 'lc' not in symbol.suffix else TdxLCMinBarReader())
            return reader.get_df(str(symbol))

        return None

    def fzline(self, symbol=None):
        """
        分钟线数据

        :param symbol: 自定义板块股票列表, 类型 list
        :return: pd.dataFrame or Bool
        """
        return self.minute(symbol, suffix=5)

    def block_new(self, name: str = None, symbol: list = None, group=False, **kwargs):
        """
        自定义板块数据操作

        :param name: 自定义板块名称
        :param symbol: 自定义板块股票列表, 类型 list
        :param group:
        :return: pd.dataFrame or Bool
        """

        return block.blocknew(name, symbol, group=group, **kwargs)

    def block(self, symbol='', group=False, **kwargs):
        """
        获取板块数据

        :param symbol:  板块文件
        :param group:   分组解析
        :return: pd.dataFrame or None
        """

        return block.block(symbol, group=group, **kwargs)


class ExtReader(ReaderBase):
    """扩展市场读取"""

    def __init__(self, tdxdir=None):
        super(ExtReader, self).__init__(tdxdir)
        self.reader = TdxExHqDailyBarReader()

    def daily(self, symbol=None):
        """
        获取扩展市场日线数据

        :return: pd.dataFrame or None
        """

        vipdoc = self.find_path(symbol=symbol, subdir='lday', suffix='day')
        return self.reader.get_df(str(vipdoc)) if vipdoc else None

    def minute(self, symbol=None):
        """
        获取扩展市场分钟线数据

        :return: pd.dataFrame or None
        """

        if not symbol:
            return None

        vipdoc = self.find_path(symbol=symbol, subdir='minline', suffix=['lc1', '1'])
        return self.reader.get_df(str(vipdoc)) if vipdoc else None

    def fzline(self, symbol=None):
        """
        获取日线数据

        :return: pd.dataFrame or None
        """

        vipdoc = self.find_path(symbol=symbol, subdir='fzline', suffix='lc5')
        return self.reader.get_df(str(vipdoc)) if symbol else None



# -*- coding: utf-8 -*-
# @Author  : BoPo
# @Time    : 2021/10/9 10:56
# @Function:


import datetime
import re
import time
from pathlib import Path

import httpx
import pandas as pd
# from py_mini_racer import py_mini_racer
from tenacity import stop_after_attempt, wait_fixed, retry

from mootdx import get_config_path
# from mootdx.consts import return_last_value
from mootdx.logger import logger

# hk_js_decode = """
# function d(t) {
#     var e, i, n, r, a, o, s, l = (arguments,
#             864e5), u = 7657, c = [], h = [], d = ~(3 << 30), f = 1 << 30,
#         p = [0, 3, 5, 6, 9, 10, 12, 15, 17, 18, 20, 23, 24, 27, 29, 30], m = Math, g = function () {
#             var l, u;
#             for (l = 0; 64 > l; l++)
#                 h[l] = m.pow(2, l),
#                 26 > l && (c[l] = v(l + 65),
#                     c[l + 26] = v(l + 97),
#                 10 > l && (c[l + 52] = v(l + 48)));
#             for (c.push("+", "/"),
#                      c = c.join(""),
#                      i = t.split(""),
#                      n = i.length,
#                      l = 0; n > l; l++)
#                 i[l] = c.indexOf(i[l]);
#             return r = {},
#                 e = o = 0,
#                 a = {},
#                 u = w([12, 6]),
#                 s = 63 ^ u[1],
#             {
#                 _1479: T,
#                 _136: _,
#                 _200: S,
#                 _139: k,
#                 _197: _mi_run
#             }["_" + u[0]] || function () {
#                 return []
#             }
#         }, v = String.fromCharCode, b = function (t) {
#             return t === {}._
#         }, N = function () {
#             var t, e;
#             for (t = y(),
#                      e = 1; ;) {
#                 if (!y())
#                     return e * (2 * t - 1);
#                 e++
#             }
#         }, y = function () {
#             var t;
#             return e >= n ? 0 : (t = i[e] & 1 << o,
#                 o++,
#             o >= 6 && (o -= 6,
#                 e++),
#                 !!t)
#         }, w = function (t, r, a) {
#             var s, l, u, c, d;
#             for (l = [],
#                      u = 0,
#                  r || (r = []),
#                  a || (a = []),
#                      s = 0; s < t.length; s++)
#                 if (c = t[s],
#                     u = 0,
#                     c) {
#                     if (e >= n)
#                         return l;
#                     if (t[s] <= 0)
#                         u = 0;
#                     else if (t[s] <= 30) {
#                         for (; d = 6 - o,
#                                    d = c > d ? d : c,
#                                    u |= (i[e] >> o & (1 << d) - 1) << t[s] - c,
#                                    o += d,
#                                o >= 6 && (o -= 6,
#                                    e++),
#                                    c -= d,
#                                    !(0 >= c);)
#                             ;
#                         r[s] && u >= h[t[s] - 1] && (u -= h[t[s]])
#                     } else
#                         u = w([30, t[s] - 30], [0, r[s]]),
#                         a[s] || (u = u[0] + u[1] * h[30]);
#                     l[s] = u
#                 } else
#                     l[s] = 0;
#             return l
#         }, x = function (t) {
#             var e, i, n;
#             for (t > 1 && (e = 0),
#                      e = 0; t > e; e++)
#                 r.d++,
#                     n = r.d % 7,
#                 (3 == n || 4 == n) && (r.d += 5 - n);
#             return i = new Date,
#                 i.setTime((u + r.d) * l),
#                 i
#         }, S = function () {
#             var t, i, a, o, l;
#             if (s >= 1)
#                 return [];
#             for (r.d = w([18], [1])[0] - 1,
#                      a = w([3, 3, 30, 6]),
#                      r.p = a[0],
#                      r.ld = a[1],
#                      r.cd = a[2],
#                      r.c = a[3],
#                      r.m = m.pow(10, r.p),
#                      r.pc = r.cd / r.m,
#                      i = [],
#                      t = 0; o = {
#                 d: 1
#             },
#                  y() && (a = w([3])[0],
#                      0 == a ? o.d = w([6])[0] : 1 == a ? (r.d = w([18])[0],
#                          o.d = 0) : o.d = a),
#                      l = {
#                          day: x(o.d)
#                      },
#                  y() && (r.ld += N()),
#                      a = w([3 * r.ld], [1]),
#                      r.cd += a[0],
#                      l.close = r.cd / r.m,
#                      i.push(l),
#                  !(e >= n) && (e != n - 1 || 63 & (r.c ^ t + 1)); t++)
#                 ;
#             return i[0].prevclose = r.pc,
#                 i
#         }, _ = function () {
#             var t, i, a, o, l, u, c, h, d, f, p;
#             if (s > 2)
#                 return [];
#             for (c = [],
#                      d = {
#                          v: "volume",
#                          p: "price",
#                          a: "avg_price"
#                      },
#                      r.d = w([18], [1])[0] - 1,
#                      h = {
#                          day: x(1)
#                      },
#                      a = w(1 > s ? [3, 3, 4, 1, 1, 1, 5] : [4, 4, 4, 1, 1, 1, 3]),
#                      t = 0; 7 > t; t++)
#                 r[["la", "lp", "lv", "tv", "rv", "zv", "pp"][t]] = a[t];
#             for (r.m = m.pow(10, r.pp),
#                      s >= 1 ? (a = w([3, 3]),
#                          r.c = a[0],
#                          a = a[1]) : (a = 5,
#                          r.c = 2),
#                      r.pc = w([6 * a])[0],
#                      h.pc = r.pc / r.m,
#                      r.cp = r.pc,
#                      r.da = 0,
#                      r.sa = r.sv = 0,
#                      t = 0; !(e >= n) && (e != n - 1 || 7 & (r.c ^ t)); t++) {
#                 for (l = {},
#                          o = {},
#                          f = r.tv ? y() : 1,
#                          i = 0; 3 > i; i++)
#                     if (p = ["v", "p", "a"][i],
#                     (f ? y() : 0) && (a = N(),
#                         r["l" + p] += a),
#                         u = "v" == p && r.rv ? y() : 1,
#                         a = w([3 * r["l" + p] + ("v" == p ? 7 * u : 0)], [!!i])[0] * (u ? 1 : 100),
#                         o[p] = a,
#                     "v" == p) {
#                         if (!(l[d[p]] = a) && (s > 1 || 241 > t) && (r.zv ? !y() : 1)) {
#                             o.p = 0;
#                             break
#                         }
#                     } else
#                         "a" == p && (r.da = (1 > s ? 0 : r.da) + o.a);
#                 r.sv += o.v,
#                     l[d.p] = (r.cp += o.p) / r.m,
#                     r.sa += o.v * r.cp,
#                     l[d.a] = b(o.a) ? t ? c[t - 1][d.a] : l[d.p] : r.sv ? ((m.floor((r.sa * (2e3 / r.m) + r.sv) / r.sv) >> 1) + r.da) / 1e3 : l[d.p] + r.da / 1e3,
#                     c.push(l)
#             }
#             return c[0].date = h.day,
#                 c[0].prevclose = h.pc,
#                 c
#         }, T = function () {
#             var t, e, i, n, a, o, l;
#             if (s >= 1)
#                 return [];
#             for (r.lv = 0,
#                      r.ld = 0,
#                      r.cd = 0,
#                      r.cv = [0, 0],
#                      r.p = w([6])[0],
#                      r.d = w([18], [1])[0] - 1,
#                      r.m = m.pow(10, r.p),
#                      a = w([3, 3]),
#                      r.md = a[0],
#                      r.mv = a[1],
#                      t = []; a = w([6]),
#                      a.length;) {
#                 if (i = {
#                     c: a[0]
#                 },
#                     n = {},
#                     i.d = 1,
#                 32 & i.c)
#                     for (; ;) {
#                         if (a = w([6])[0],
#                         63 == (16 | a)) {
#                             l = 16 & a ? "x" : "u",
#                                 a = w([3, 3]),
#                                 i[l + "_d"] = a[0] + r.md,
#                                 i[l + "_v"] = a[1] + r.mv;
#                             break
#                         }
#                         if (32 & a) {
#                             o = 8 & a ? "d" : "v",
#                                 l = 16 & a ? "x" : "u",
#                                 i[l + "_" + o] = (7 & a) + r["m" + o];
#                             break
#                         }
#                         if (o = 15 & a,
#                             0 == o ? i.d = w([6])[0] : 1 == o ? (r.d = o = w([18])[0],
#                                 i.d = 0) : i.d = o,
#                             !(16 & a))
#                             break
#                     }
#                 n.date = x(i.d);
#                 for (o in {
#                     v: 0,
#                     d: 0
#                 })
#                     b(i["x_" + o]) || (r["l" + o] = i["x_" + o]),
#                     b(i["u_" + o]) && (i["u_" + o] = r["l" + o]);
#                 for (i.l_l = [i.u_d, i.u_d, i.u_d, i.u_d, i.u_v],
#                          l = p[15 & i.c],
#                      1 & i.u_v && (l = 31 - l),
#                      16 & i.c && (i.l_l[4] += 2),
#                          e = 0; 5 > e; e++)
#                     l & 1 << 4 - e && i.l_l[e]++,
#                         i.l_l[e] *= 3;
#                 i.d_v = w(i.l_l, [1, 0, 0, 1, 1], [0, 0, 0, 0, 1]),
#                     o = r.cd + i.d_v[0],
#                     n.open = o / r.m,
#                     n.high = (o + i.d_v[1]) / r.m,
#                     n.low = (o - i.d_v[2]) / r.m,
#                     n.close = (o + i.d_v[3]) / r.m,
#                     a = i.d_v[4],
#                 "number" == typeof a && (a = [a, a >= 0 ? 0 : -1]),
#                     r.cd = o + i.d_v[3],
#                     l = r.cv[0] + a[0],
#                     r.cv = [l & d, r.cv[1] + a[1] + !!((r.cv[0] & d) + (a[0] & d) & f)],
#                     n.volume = (r.cv[0] & f - 1) + r.cv[1] * f,
#                     t.push(n)
#             }
#             return t
#         }, k = function () {
#             var t, e, i, n;
#             if (s > 1)
#                 return [];
#             for (r.l = 0,
#                      n = -1,
#                      r.d = w([18])[0] - 1,
#                      i = w([18])[0]; r.d < i;)
#                 e = x(1),
#                     0 >= n ? (y() && (r.l += N()),
#                         n = w([3 * r.l], [0])[0] + 1,
#                     t || (t = [e],
#                         n--)) : t.push(e),
#                     n--;
#             return t
#         };
#     return _mi_run = function () {
#         var t, i, a, o;
#         if (s >= 1)
#             return [];
#         for (r.f = w([6])[0],
#                  r.c = w([6])[0],
#                  a = [],
#                  r.dv = [],
#                  r.dl = [],
#                  t = 0; t < r.f; t++)
#             r.dv[t] = 0,
#                 r.dl[t] = 0;
#         for (t = 0; !(e >= n) && (e != n - 1 || 7 & (r.c ^ t)); t++) {
#             for (o = [],
#                      i = 0; i < r.f; i++)
#                 y() && (r.dl[i] += N()),
#                     r.dv[i] += w([3 * r.dl[i]], [1])[0],
#                     o[i] = r.dv[i];
#             a.push(o)
#         }
#         return a
#     }
#         ,
#         g()()
# }
# """
#
#
# @retry(wait=wait_fixed(2), retry_error_callback=return_last_value, stop=stop_after_attempt(5))
# def holiday2(date=False, save=True) -> pd.DataFrame:
#     """ 交易日历-历史数据
#     :return: 交易日历
#     :rtype: pandas.DataFrame
#     """
#     cache_file = get_config_path('holiday.csv')
#     temp_df = None
#
#     if Path(cache_file).exists():
#         df = pd.read_csv(cache_file)
#         if not df.loc[df['year'] == datetime.datetime.now().year].empty:
#             temp_df = df
#
#     if type(temp_df) is not pd.DataFrame:
#         client = httpx.Client(verify=False)
#
#         url = "https://finance.sina.com.cn/realstock/company/klc_td_sh.txt"
#         res = client.get(url)
#
#         js_code = py_mini_racer.MiniRacer()
#         js_code.eval(hk_js_decode)
#
#         # 执行js解密代码
#         dict_list = js_code.call("d", res.text.split("=")[1].split(";")[0].replace('"', ""))
#
#         temp_df = pd.DataFrame(dict_list)
#         temp_df.columns = ["date"]
#         temp_df["date"] = pd.to_datetime(temp_df["date"]).dt.date
#
#         temp_list = temp_df["date"].to_list()
#         temp_list.append(datetime.date(1992, 5, 4))  # 是交易日但是交易日历缺失该日期
#         temp_list.sort()
#
#         temp_df = pd.DataFrame(temp_list, columns=['date'])
#         temp_df['year'] = pd.DatetimeIndex(temp_df['date']).year
#
#         # 保存缓存
#         save and temp_df.to_csv(cache_file)
#
#     if date:
#
#         try:
#             date = datetime.datetime.strptime(date, "%Y-%m-%d").date()
#         except ValueError:
#             date = datetime.datetime.now().date()
#
#         return temp_df.loc[temp_df['date'] == date].all().any()
#
#     return temp_df


def holiday(date=None, format_=None, country=None):
    cache_file = get_config_path('holiday.plk')

    format_ = format_ if format_ else "%Y-%m-%d"
    country = country if country else '中国'

    try:
        if date:
            date = datetime.datetime.strptime(date, format_).date()
        else:
            date = datetime.datetime.now().date()
    except ValueError as ex:
        logger.exception("日期或者日期格式错误!")
        raise ex

    if date.weekday() >= 5:
        return True

    if Path(cache_file).exists() and time.localtime(Path(cache_file).stat().st_mtime).tm_year == time.localtime(time.time()).tm_year:
        df = pd.read_pickle(cache_file)
    else:
        res = httpx.get('https://www.tdx.com.cn/url/holiday/')
        ret = re.findall(r'<textarea id="data" style="display:none;">([\s\w\d\W]+)</textarea>', res.text, re.M)[0].strip()
        day = [d.split('|')[:4] for d in ret.split('\n')]

        df = pd.DataFrame(day, columns=['日期', '节日', '国家', '交易所'], dtype=str)
        df.index = pd.to_datetime(df['日期'].astype("str"), format='%Y%m%d')
        df.to_pickle(cache_file)

    if country not in list(set(df['国家'].values)):
        raise ValueError(f'没有该国家`{country}`的交易日数据')

    df = df[df['国家'] == country]
    return not df[df.index.isin([date])].empty



import math

import pandas
import pandas as pd
from pytdx.exhq import TdxExHq_API
from pytdx.hq import TdxHq_API
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import retry_if_result
from tenacity import stop_after_attempt
from tqdm import tqdm

from mootdx import config
from mootdx import server
from mootdx.consts import MARKET_SH, return_last_value
from mootdx.logger import logger
from mootdx.utils import get_stock_market
from mootdx.utils import get_stock_markets
from mootdx.utils import get_frequency
from mootdx.utils import to_data


class Quotes(object):
    @staticmethod
    def factory(market='std', **kwargs):
        """
        股票市场 工厂方法

        :param market:  std 股票市场, ext 扩展市场， 默认股票市场
        :param kwargs:  可变参数
        :return: object
        """
        if market == 'ext':
            return ExtQuotes(**kwargs)

        return StdQuotes(**kwargs)


class BaseQuotes(object):
    client = None
    bestip = None

    def __init__(self, bestip: bool = False, timeout: int = None, quiet=False, **kwargs) -> None:

        quiet and logger.remove()

        logger.debug(f'bestip=>{bestip}')
        bestip and server.bestip()

        logger.debug(f'timeout=>{timeout}')
        self.timeout = timeout if timeout else 15

        logger.debug('config.setup()')
        config.setup()

    def __del__(self):
        logger.debug('call __del__')
        self.close()

    def reconnect(self):
        if self.closed:
            logger.debug('服务器连接已断开，正进行重新连接...')
            self.client.connect(*self.bestip)

    def close(self):
        logger.debug('close')

        hasattr(self.client, 'close') and self.client.close()

    @property
    def closed(self) -> bool:
        if not hasattr(self.client.client, '_closed') or getattr(self.client.client, '_closed'):
            return True

        return False

    def pool(self):
        ...


instance: BaseQuotes


def check_empty(value):
    """
    重试判断函数

    :param value: 要判断的值
    :return:
    """
    _empty = value.all().empty if isinstance(value, pd.DataFrame) else not value

    # 判断状态空，则重连接
    if instance and _empty:
        logger.info('重新连接服务器')
        instance.client.connect(*instance.bestip)

    return _empty


class StdQuotes(BaseQuotes):
    """
    股票市场实时行情 """

    def __init__(self, bestip=False, timeout=15, **kwargs):
        """ 构造函数

        :param bestip:  最佳 IP
        :param timeout: 超时时间
        :param kwargs:  可变参数
        """

        super(StdQuotes, self).__init__(bestip=bestip, timeout=timeout, **kwargs)

        try:
            config.get('SERVER').get('HQ')[0]
        except ValueError as ex:
            logger.warning(ex)
        finally:
            default = config.get('SERVER').get('HQ')[0][1:]
            self.bestip = config.get('BESTIP').get('HQ', default)

        if kwargs.get('quiet'):
            del kwargs['quiet']

        # ip_pool = AvailableIPPool(TdxHq_API, config.get('SERVER').get('HQ')[:5])
        # self.client = TdxHqPool_API(TdxHq_API, ip_pool)

        self.client = TdxHq_API(raise_exception=False, **kwargs)
        self.client.connect(*self.bestip)

        global instance
        instance = self

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def traffic(self):
        return self.client.get_traffic_stats()

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def quotes(self, symbol=None, **kwargs):
        """
        获取实时日行情数据

        :param symbol: 股票代码
        :return: pd.dataFrame or None
        """

        if not symbol:
            return to_data(None)

        if type(symbol) is str:
            symbol = [symbol]

        symbol = get_stock_markets(symbol)
        result = self.client.get_security_quotes(symbol)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def bars(self, symbol='000001', frequency=9, start=0, offset=100, **kwargs):
        """
        获取实时日K线数据

        :param symbol: 股票代码
        :param frequency: 数据频次
        :param start: 开始位置
        :param offset: 每次获取条数
        :return: pd.dataFrame or None
        """
        frequency = get_frequency(frequency)
        market = get_stock_market(symbol)
        offset = 800 if int(offset) >= 800 else offset
        result = self.client.get_security_bars(int(frequency), int(market), str(symbol), int(start), int(offset))

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def stock_count(self, market=MARKET_SH):
        """
        获取市场股票数量

        :param market: 股票市场代码 sh 上海， sz 深圳
        :return: pd.dataFrame or None
        """

        result = self.client.get_security_count(market=market)

        return result

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def stocks(self, market=MARKET_SH):
        """
        获取股票列表

        :param market: 股票市场
        :return:
        """

        counts = self.client.get_security_count(market=market)
        stocks = None

        for start in tqdm(range(0, counts, 1000)):
            result = self.client.get_security_list(market=market, start=start)
            stocks = (pandas.concat([stocks, to_data(result)], ignore_index=True) if start > 1 else to_data(result))

        return stocks

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def index_bars(self, symbol='000001', frequency=9, start=0, offset=100, **kwargs):
        """
        获取指数k线

        :param symbol: 股票代码
        :param frequency: 数据频次
        :param start: 开始位置
        :param offset: 获取数量
        :return:
        """

        frequency = get_frequency(frequency)

        market = get_stock_market(symbol)
        result = self.client.get_index_bars(frequency=frequency, market=market, code=symbol, start=start, count=offset)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def minute(self, symbol='', **kwargs):
        """
        获取实时分时数据

        :param symbol: 股票代码
        :return: pd.DataFrame
        """

        market = get_stock_market(symbol)
        result = self.client.get_minute_time_data(market=market, code=symbol)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def minutes(self, symbol='', date='20191023', **kwargs):
        """
        分时历史数据

        :param symbol:  股票代码
        :param date:    查询日期
        :return: pd.dataFrame or None
        """

        market = get_stock_market(symbol)
        result = self.client.get_history_minute_time_data(market=market, code=symbol, date=date)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def transaction(self, symbol='', start=0, offset=10, **kwargs):
        """
        查询分笔成交

        :param symbol:  股票代码
        :param start:   起始位置
        :param offset:  结束位置
        :return: pd.dataFrame or None
        """

        market = get_stock_market(symbol)
        result = self.client.get_transaction_data(int(market), symbol, start, offset)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def transactions(self, symbol='', start=0, offset=10, date='20170209', **kwargs):
        """
        查询历史分笔成交

        :param symbol:  股票代码
        :param start:   起始位置
        :param offset:  获取数量
        :param date:    查询日期
        :return: pd.dataFrame or None
        """

        market = get_stock_market(symbol, string=False)
        result = self.client.get_history_transaction_data(market=market, code=symbol, start=start, count=offset, date=int(date))

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def F10C(self, symbol=''):  # noqa
        """
        查询公司信息目录

        :param symbol: 股票代码
        :return: pd.dataFrame or None
        """

        market = int(get_stock_market(symbol))
        result = self.client.get_company_info_category(market, symbol)

        return result

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def F10(self, symbol='', name=''):  # noqa
        """
        读取公司信息详情

        :param name: 公司 F10 标题
        :param symbol: 股票代码
        :return: pd.dataFrame or None
        """

        result = {}
        market = int(get_stock_market(symbol, string=False))

        category = self.client.get_company_info_category(market, symbol)

        if not category:
            return None

        if name:
            for x in category:
                if x['name'] == name:
                    return self.client.get_company_info_content(
                        market=market,
                        code=symbol,
                        filename=x['filename'],
                        start=x['start'],
                        length=x['length'],
                    )

        for x in category:
            result[x['name']] = self.client.get_company_info_content(market=market, code=symbol, filename=x['filename'], start=x['start'], length=x['length'])

        return result

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def xdxr(self, symbol='', **kwargs):
        """
        读取除权除息信息

        :param symbol: 股票代码
        :return: pd.dataFrame or None
        """

        market = get_stock_market(symbol)
        result = self.client.get_xdxr_info(int(market), symbol)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def finance(self, symbol='000001', **kwargs):
        """
        读取财务信息

        :param symbol: 股票代码
        :return:
        """

        market = get_stock_market(symbol)
        result = self.client.get_finance_info(market=market, code=symbol)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def k(self, symbol='', begin=None, end=None):
        """
        读取k线信息

        :param symbol:  股票代码
        :param begin:   开始日期
        :param end:     截止日期
        :return: pd.dataFrame or None
        """

        result = self.get_k_data(symbol, begin, end)
        return result

    def ohlc(self, **kwargs):
        return self.k(**kwargs)

    def get_k_data(self, code, start_date, end_date):

        # !! todo
        data = pd.concat([self.client.to_df(self.client.get_security_bars(9, get_stock_market(code), code, (9 - i) * 800, 800)) for i in range(10)], axis=0)
        data = data.assign(date=data['datetime'].apply(lambda x: str(x)[0:10])).assign(code=str(code))
        data = data.set_index('date', drop=False, inplace=False)
        data = data.drop(['year', 'month', 'day', 'hour', 'minute', 'datetime'], axis=1)[start_date:end_date]
        data = data.assign(date=data['date'].apply(lambda x: str(x)[0:10]))

        return data

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def index(self, symbol='000001', market=MARKET_SH, frequency=9, start=1, offset=2, **kwargs):
        """
        获取指数k线

        K线种类:
        - 0 5分钟K线
        - 1 15分钟K线
        - 2 30分钟K线
        - 3 1小时K线
        - 4 日K线
        - 5 周K线
        - 6 月K线
        - 7 1分钟
        - 8 1分钟K线
        - 9 日K线
        - 10 季K线
        - 11 年K线

        :param symbol:      股票代码
        :param frequency:   数据频次
        :param market:      证券市场
        :param start:       开始位置
        :param offset:      每次获取条数
        :return: pd.dataFrame or None
        """

        result = self.client.get_index_bars(int(frequency), int(market), str(symbol), int(start), int(offset))

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def block(self, tofile='block.dat', **kwargs):
        """
        获取证券板块信息

        :param tofile: 保存文件
        :return: pd.dataFrame or None
        """

        result = self.client.get_and_parse_block_info(tofile)
        return to_data(result, **kwargs)


class ExtQuotes(BaseQuotes):
    """ 扩展市场实时行情 """

    bestip = ('112.74.214.43', 7727)

    def __init__(self, bestip=False, timeout=15, **kwargs):
        """
        构造函数

        :param bestip:  最优服务器IP
        :param timeout: 超时时间
        :param kwargs:  可变参数
        """
        super(ExtQuotes, self).__init__(bestip=bestip, timeout=timeout, **kwargs)

        logger.warning('目前扩展市场行情接口已经失效, 后期有望修复.')

        try:
            config.get('SERVER').get('EX')[0]
        except ValueError:
            server.bestip()
        finally:
            default = config.get('SERVER').get('EX')[0]
            self.bestip = config.get('BESTIP').get('EX', default)

        if kwargs.get('quiet'):
            del kwargs['quiet']

        self.client = TdxExHq_API(**kwargs)
        self.client.connect(*self.bestip)

    @staticmethod
    def validate(market, symbol):
        """
        验证股票市场

        :param market: 股票市场
        :param symbol: 股票代码
        :return: tuple
        """

        if not market:
            if len(symbol.split('#')) > 1:
                market = symbol.split('#')[0]
                symbol = symbol.split('#')[1]

        if not market:
            raise ValueError('市场参数错误, 市场参数不能为空.')

        return int(market), symbol

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def markets(self, **kwargs):
        """
        获取实时市场列表

        :return: pd.dataFrame or None
        """

        result = self.client.get_markets()
        return to_data(result, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def instrument(self, start=0, offset=100, **kwargs):
        """
        查询代码列表

        :param start:   开始位置
        :param offset:  获取数量
        :return:
        """

        result = self.client.get_instrument_info(start=start, count=offset)
        return to_data(result, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def instrument_count(self):
        """
        市场商品数量

        :return:
        """

        result = self.client.get_instrument_count()

        return result

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def instruments(self, **kwargs):
        """
        查询所有代码列表

        :return:
        """

        result = []

        count = self.client.get_instrument_count()
        pages = math.ceil(count / 100)

        for page in tqdm(range(0, pages)):
            result += self.client.get_instrument_info(page * 100, 100)

        return to_data(result, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def quote(self, market='', symbol='', **kwargs):
        """
        查询五档行情

        :param market: 市场ID
        :param symbol: 证券代码
        :return:
        """

        market, symbol = self.validate(market, symbol)
        result = self.client.get_instrument_quote(market, symbol)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def minute(self, market='', symbol='', **kwargs):
        """
        查询分时行情

        :param market: 市场ID
        :param symbol: 证券代码
        :return:
        """

        market, symbol = self.validate(market, symbol)
        result = self.client.get_minute_time_data(market, symbol)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def minutes(self, market=None, symbol='', date='', **kwargs):
        """
        查询历史分时行情

        :param market:  市场ID
        :param symbol:  证券代码
        :param date:    查询日期
        :return:
        """

        market, symbol = self.validate(market, symbol)
        result = self.client.get_history_minute_time_data(market, symbol, date)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def bars(self, frequency='', market='', symbol='', start=0, offset=100, **kwargs):
        """
        查询k线数据

        :param frequency: 数据频次, K线周期
        :param market: 市场ID
        :param symbol: 证券代码
        :param start:  起始位置
        :param offset: 获取数量
        :return:
        """

        frequency = get_frequency(frequency)

        market, symbol = self.validate(market, symbol)
        result = self.client.get_instrument_bars(category=frequency, market=market, code=symbol, start=start, count=offset)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def transaction(self, market=None, symbol='', start=0, offset=1800, **kwargs):
        """
        查询分笔成交

        :param market: 市场ID
        :param symbol: 证券代码
        :param start:  开始位置
        :param offset: 获取数量
        :return:
        """

        market, symbol = self.validate(market, symbol)
        result = self.client.get_transaction_data(market=market, code=symbol, start=start, count=offset)

        return to_data(result, symbol=symbol, **kwargs)

    @retry(stop=stop_after_attempt(3), retry_error_callback=return_last_value, retry=(retry_if_exception_type() | retry_if_result(check_empty)))
    def transactions(self, market=None, symbol='', date='', start=0, offset=1800, **kwargs):
        """
        查询历史分笔成交

        :param market:  市场ID
        :param symbol:  证券代码
        :param date:    查询日期
        :param start:   开始位置
        :param offset:  获取数量
        :return:
        """

        market, symbol = self.validate(market, symbol)
        result = self.client.get_history_transaction_data(market=market, code=symbol, date=int(date), start=start, count=offset)

        return to_data(result, symbol=symbol, **kwargs)
